/* Functions to compute edginess.

   For the article
   "Singularity Detection in Images Using Dual Local Autocovariance"
   by  Wojciech Czaja and Mladen Victor Wickerhauser.

   Programmer: M.V.Wickerhauser <victor@math.wustl.edu>
   Date: 22 July 2001

 */

#include <assert.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#include "edginess.h"

#define WAY_TOO_BIG	4097
#ifndef PI
# define PI		3.1415926535897973
#endif


/* Allocate a Grid data structure and its samples array. */
Grid
newGrid (
	 int r0,
	 int r1,
	 int c0,
	 int c1,
	 const real *in)	/* if non-NULL, copy these values */
{
  Grid new;
  int k;

  assert(r0<=r1);  assert(c0<=c1);
  new = (Grid)malloc(sizeof(struct Grid)); assert(new);
  new->r0 = r0;   new->r1=r1;   new->c0=c0;   new->c1=c1;
  new->rows = 1 + r1 - r0;      new->cols = 1 + c1 - c0;
  new->samples = (real **) malloc ( new->rows * sizeof(real *) );
  assert(new->samples);
  for(k=0; k<new->rows; k++) {
    new->samples[k] = (real *) calloc ( new->cols, sizeof(real) );
    assert(new->samples[k]);
    new->samples[k] -= c0;	/* `samples[k][c]' now valid for c0<=c<=c1 */
  }
  new->samples -= r0;		/* `samples[r][*]' now valid for r0<=r<=r1 */

  if(in) {			/* assume the data is stored row-by-row */
    int i,j;
    for(i=r0;i<=r1; i++)
      for(j=c0; j<=c1; j++)
	new->samples[i][j] = *in++;
  } /* ... otherwise, leave `new->samples[][]' filled with zeroes. */
  return new;
}

/* Return a new Grid data structure with samples array of
   the same dimensions as the input Grid, but all zeroes. */
Grid
likeGrid (
	 Grid in )
{
  Grid new;

  assert(in);
  new = newGrid( in->r0,in->r1, in->c0,in->c1, 0);
  return new;
}

/* Return a new Grid data structure with samples array of
   the same dimensions as the input Grid, but all zeroes. */
Grid
copyGrid (
	  Grid out,		/* If NULL, allocate a new one */
	  Grid in )		/* Copy samples from here */
{
  int i,j;

  assert(in);
  if(!out)  out = likeGrid( in );
  else {
    assert(out->r0 == in->r0);    assert(out->r1 == in->r1);
    assert(out->c0 == in->c0);    assert(out->c1 == in->c1);
  }
  for ( i = in->r0  ;  i <= in->r1  ; i++ )
    for ( j = in->c0  ;  j <= in->c1  ; j++ )
      out->samples[i][j] = in->samples[i][j];
  return out;
}


/* Deallocate a Grid data structure and its samples array. */
void
freeGrid ( Grid in )
{
  int k;
  if(in) {
    for(k=in->r0; k<=in->r1; k++ )
      free( in->samples[k] + in->c0 );
    free( in->samples + in->r0 );
    free(in);
  }
  return;
}


/* Return the sample at the requested grid point.
   Return zero for off-grid indices.  */
real
sample( Grid g, int row, int col )
{
  real out=0;
  if ( row >= g->r0  &&  col >= g->c0  &&  row <= g->r1  &&  col <= g->c1 )
    out = g->samples[row][col];
  return out;
}

/* Insert the value `x' as a sample at the requested grid point.
   Do nothing at off-grid indices.  */
void
insert( Grid g, int row, int col, real x )
{
  if ( row >= g->r0  &&  col >= g->c0  &&  row <= g->r1  &&  col <= g->c1 )
    g->samples[row][col] = x;
  return;
}
    
/* Read a PGM file into a Grid data structure, with no error checking.
   The file's header must be exactly
	P5
	<width> <height>
	255
	<byte-data>
*/
Grid
readPGM ( FILE *pgmfile )
{
  int i,j, cols,rows;
  Grid out;
  unsigned char *buf, *bp;

  assert(pgmfile);

  fscanf( pgmfile, "P5\n%d%d\n255\n", &cols,&rows );
  assert(cols>0);  assert(cols<WAY_TOO_BIG);
  assert(rows>0);  assert(rows<WAY_TOO_BIG);

  bp = buf = (unsigned char *)malloc(rows*cols);  assert(buf);
  fread(buf, 1, rows*cols, pgmfile);

  out = newGrid( 0, rows-1, 0, cols-1, 0 );
  
  for(i=0; i<rows; i++)
    for(j=0; j<cols; j++)
      out->samples[i][j] = (real) *bp++;

  free(buf);
  return out;
}

/* Write a PGM file from data in a Grid data structure,
   assumed normalized. The file's header will be exactly
	P5
	<cols> <rows>
	255
	<byte-data>
*/
void
writePGM ( FILE *pgmfile, Grid in )
{
  int i,j;
  assert(pgmfile);

  fprintf(pgmfile,"P5\n%d %d\n255\n", in->cols,in->rows); /* Header */
  for(i=in->r0; i<=in->r1; i++)
    for(j=in->c0; j<=in->c1; j++)
      putc((int)(0.5+in->samples[i][j]), pgmfile); /* Byte-data */

  return;
}

/* Scale and possibly shift the sample values in a Grid to fill the
   dynamic range of 8-bit grayscale display devices.  If all values
   are nonnegative, then scale so that the largest value is 255.
   If there are some negative values, shift the zero value to 128
   and scale to fit within the range [0,255].
   */
real
normalize(Grid g)
{
  int i,j;
  real x, xmax, xmin, amax, norm;

  assert(g);  assert(g->samples);

  /* Find maximum and minimum */
  xmax = xmin = g->samples[g->r0][g->c0];
  for ( i = g->r0  ;  i <= g->r1  ;  i++ )
    for ( j = g->c0  ;  j <= g->c1  ;  j++ ) {
      x = g->samples[i][j];
      if( x>xmax ) xmax = x;
      else
	if( x<xmin ) xmin = x;
    }
  if ( xmin >= 0 )		/* All sample values are nonnegative. */
    if ( xmax == 0 )
      amax = 0;			/* Do nothing; the grid is all zeroes */
    else {
      amax = xmax;
      norm = 255/xmax;
      for ( i = g->r0  ;  i <= g->r1  ;  i++ )
	for ( j = g->c0  ;  j <= g->c1  ;  j++ )
	  g->samples[i][j] *= norm;
    }
  else {			/* Some samples are negative */
    amax = ( xmax > -xmin ) ? xmax : -xmin ;
    norm = 127/amax;
    amax = -amax;		/* for return, to indicate negatives */
    for ( i = g->r0  ;  i <= g->r1  ;  i++ )
      for ( j = g->c0  ;  j <= g->c1  ;  j++ ) 
	g->samples[i][j] = norm * g->samples[i][j] + 128;
  }
  return amax;
}


/* Extract a subgrid around the gridpoint (i0,j0). */
void
sharp( Grid out, Grid in, int i0, int j0 )
{
  int i,j;
  for ( i = out->r0  ;  i <= out->r1  ;  i++ )
    for ( j = out->c0  ;  j <= out->c1  ;  j++ )
      insert( out, i,j,  sample(in,i0+i,j0+j) );
  return; 
}


/* Extract a subgrid around the gridpoint (i0,j0),
   using a sampled bump to weight the samples.
 */
void
smooth( Grid out, Grid in, int i0, int j0, Grid bump )
{
  int i,j;
  for ( i = out->r0  ;  i <= out->r1  ; i++ )
    for ( j = out->c0  ;  j <= out->c1  ; j++ )
      insert( out, i,j,  sample(bump,i,j)*sample(in,i0+i,j0+j) );
  return; 
}

/* Generate a cut-off Gaussian bump function on the square grid
   [-e,e] x [-e,e], centered at 0.
 */
Grid
gaussian( int e )
{
  Grid bump;
  int i,j;

  assert(e>=0);
  bump = newGrid( -e,e, -e,e, 0);
  if ( e > 0 )
    for ( i = -e  ;  i <= e  ;  i++ )
      for ( j = -e  ;  j <= e  ;  j++ )
	bump->samples[i][j] = exp( -PI*(i*i+j*j)/(double)(e*e));
  else bump->samples[0][0] = 1;
  return bump; 
}


/* Compute the reciprocal "edginess" from the eigenvalues.  Return values are
   in the range [0,1], with 0 indicating greatest edginess and 1 indicating
   least edginess.  This is intended, after normalization, for write-black
   display devices, where 0 is blackest.  */
real
edginess ( const real evs[2] )
{
  real out=1;
  if(evs[0]>0)
    out = evs[1]/evs[0];
  return out;
}

/* Return the eigenvalues lambda[0],lambda[1]
   of a symmetric 2x2 matrix. */
void
eigenvalues( real lambda[2], real symmat[2][2] )
{
  real a,b,c,d;
  assert(lambda);
  a = symmat[0][0]+symmat[1][1];
  b = symmat[0][0]-symmat[1][1];
  c = symmat[0][1]+symmat[1][0]; /* == 2*symmat[0][1] */
  d = sqrt(b*b+c*c);
  lambda[0] = 0.5 * (a + d);
  lambda[1] = 0.5 * (a - d);
  return;
}


/* First moment kernel, less a factor of i/(2*pi) */
static real
mu1( int n )
{
  real out=0;			/* value at `n==0' */
  if(n)
    out = n&1 ? -1/(real)n : 1/(real)n;
  return out;
}

/* Second moment kernel, less a factor of 1/(2*pi^2) */
static real
mu2( int n )
{
  real out = PI*PI/6;		/* value at `n==0', times 2*pi^2 */
  if(n)
    out = n&1 ? -1/(real)(n*n) : 1/(real)(n*n);
  return out;
}

/* Return the 2x2 dual local autocovariance matrix,
   a second moment matrix obtained from a subgrid.
 */
void
momenta( real out[2][2], Grid in )
{
  int n,nn,m,mm;
  real sum;
  assert(out);
  assert(in);

  /* E11: */
  sum = 0;
  for ( m = in->r0  ;  m <= in->r1  ;  m++ )
    for ( mm = in->r0  ;  mm <= in->r1  ;  mm++ )
      for ( n = in->c0  ;  n <= in->c1  ;   n++ )
	sum += in->samples[m][n] * in->samples[mm][n] * mu2(m-mm) ;
  out[0][0] = sum / (2*PI*PI);	/* Replace factor missing from mu2() */

  /* E22: */
  sum = 0;
  for ( m = in->r0  ;  m <= in->r1  ;  m++ )
    for ( n = in->c0  ;  n <= in->c1  ;  n++ )
      for ( nn = in->c0  ;  nn <= in->c1  ;  nn++ )
	sum += in->samples[m][n] * in->samples[m][nn] * mu2(n-nn) ;
  out[1][1] = sum / (2*PI*PI);	/* Replace factor missing from mu2() */


  /* E21 == E12: */
  sum = 0;
  for ( m = in->r0  ;  m <= in->r1  ;  m++ )
    for ( mm = in->r0  ;  mm <= in->r1  ;  mm++ )
      for ( n = in->c0  ;  n <= in->c1  ;  n++ )
	for ( nn = in->c0  ;  nn <= in->c1  ;  nn++ )
	  sum += in->samples[m][n] * in->samples[mm][nn] * mu1(m-mm)*mu1(n-nn);
  out[1][0] = out[0][1]		/* x*y moment equals y*x moment */
    = -sum / (4*PI*PI);		/* Replace missing i/(2*PI) for each mu1() */

  return;
}

void
printGrid( Grid in, const char *message )
{
  int i,j;
  puts(message);
  printf("  Index range [%d,%d]x[%d,%d] (%d rows by %d columns)\n",
	 in->r0,in->r1, in->c0,in->c1, in->rows,in->cols );
  if( in->rows < WAY_TOO_BIG  && in->cols < WAY_TOO_BIG )
    for ( i = in->r0  ;  i <= in->r1  ;  i++ ) {
      for ( j = in->c0  ;  j <= in->c1  ;  j++ )
	printf("  %12f", in->samples[i][j] );
      putchar('\n');
    }
  else
    puts("  [way too big to print]");
  return;
}

void
print2x2( real in[2][2], const char *message )
{
  int i,j;
  puts(message);
  for(i=0; i<2; i++) {
    for(j=0; j<2; j++)
      printf("  %12f",in[i][j]);
    putchar('\n');
  }
  return;
}
