/* Tests of functions to compute edginess.

   For the article
   "Singularity Detection in Images Using Dual Local Autocovariance"
   by  Wojciech Czaja and Mladen Victor Wickerhauser.

   Programmer: M.V.Wickerhauser <victor@math.wustl.edu>
   Date: 22 July 2001

 */

#include <assert.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#include "edginess.h"

int
main(void)			/* test some functions */
{
  FILE *testin;			/* for a small test PGM file  */
  Grid g;			/* the samples that define `testin' */


  /*  newGrid(), sample(), insert() */
  {
    real x[2000];		/* 40 rows of 50 columns */
    int k;

    /* Make sure there's an edge between rows 19 and 20: */
    for(k=0; k<1000; k++)	x[k] = k%50;
    for(k=1000; k<2000; k++)	x[k] = 110-2*(k%50);

    /* Reindex the edge between rows 9 and 10: */
    g = newGrid ( -10,29, -25,24, x );    assert(g);

    assert( sample(g, 0,0)== 25);
    assert( sample(g, 10,-20)== 100);
    assert( sample(g, 10,-26) == 0); /* off grid */
    assert( sample(g, 100,0) == 0); /* off grid */

    /* Insert recognizable values that don't interfere with later tests: */
    insert( g,  0,-5,  66 );	/* upper left */
    insert( g,  0, 0,  77 );	/* upper right */
    insert( g,  3, 0,  88 );	/* lower right */
    insert( g,  3,-5,  99 );	/* lower left */
    insert( g, 30,-25, -5 );	/* invalid */
    insert( g, -10,25, -7 );	/* invalid */

    assert( sample( g,  0,-5 ) == 66 );
    assert( sample( g,  0, 0 ) == 77 );
    assert( sample( g,  3, 0 ) == 88 );
    assert( sample( g,  3,-5 ) == 99 );
    assert( sample( g, 30,-25) ==  0 );	/* invalid */
    assert( sample( g,-10,25 ) ==  0 );	/* invalid */
  }

  /* writePGM() */
  {
    testin = fopen("test.pgm","wb");  assert(testin);
    writePGM( testin, g );
    fclose(testin);
    puts("Wrote test.pgm --- 40 rows [-10:29] by 50 columns [-25:24].");

    testin = fopen("test.pgm","rb");  assert(testin);
  }

  /* normalize(), readPGM() */
  {
    real smax;
    Grid copy;
    FILE *outfile;

    outfile = fopen("testnorm.pgm","wb");
    copy = readPGM(testin);
    smax = normalize( copy );
    assert(smax == 110);
    writePGM( outfile, copy );
    fclose( outfile );
    puts("Wrote testnorm.pgm --- Normalized copy of test.pgm.");

    freeGrid ( copy );
  }

  /* print2x2(), eigenvalues() */
  {
    real mat[2][2]={{1,2},{3,4}};
    real lambda[2];

    print2x2(mat, "Matrix should be {{1,2},{3,4}}:");
    mat[0][0] =  2;    mat[0][1] = -1;
    mat[1][0] = -1;    mat[1][1] =  2;
    eigenvalues ( lambda, mat );
    assert( lambda[0] == 3 );	/* larger */
    assert( lambda[1] == 1 );	/* smaller */
  }

  /* printGrid(), sharp() */
  {
    Grid sub;
    sub = newGrid(0,3, -5,0, 0);
    sharp( sub, g, 0,0 );
    printGrid( sub, "4x6 portion of g");
    freeGrid(sub);
  }

  /* gaussian(), smooth(), likeGrid(), copyGrid() */
  {
    Grid bump, outg, subg;
    FILE *outfile;

    /* Trivial case: */
    bump = gaussian(0);    assert(bump);    assert(sample(bump,0,0)==1);
    freeGrid(bump);

    /* Less trivial bump, used in smooth(): */
    bump = gaussian(5);   assert(sample(bump,0,0)==1);

    outfile = fopen("testbump.pgm","wb");
    outg = copyGrid( 0, bump );
    normalize( outg );
    writePGM( outfile, outg );
    fclose(outfile);
    puts("Wrote testbump.pgm --- density plot of 11 x 11 Gaussian.");

    subg = likeGrid( bump );
    smooth( subg, g, 0,0, bump );
    outfile = fopen("testsmoo.pgm","wb");
    copyGrid( outg, subg );
    normalize ( outg );
    writePGM( outfile, outg );
    fclose(outfile);
    puts("Wrote testsmoo.pgm --- density plot of smoothed `test.pgm'.");

    freeGrid(outg);
    freeGrid(subg);
    freeGrid(bump);
  }

  /* edginess() */
  {
    real y[2];

    y[0]=2; y[1]=1;	assert ( edginess(y) == 0.5);
    y[0]=10; y[1]=0;	assert ( edginess(y) == 0 );
  }

  /* momenta() */
  {
    FILE *outfile;
    Grid bump, subg;
    real mat[2][2]={{0,0},{0,0}}, y[2]={0,0};

    bump = gaussian(5);
    subg = likeGrid(bump);
    
    /* LEFT, big positive horizontal edge */
    /* Plot the locale: */
    sharp ( subg, g,  10,-19 );    normalize ( subg );
    outfile = fopen ( "testlt.pgm", "wb");  assert(outfile);
    writePGM ( outfile, subg );   fclose ( outfile );
    puts("Wrote testlt.pgm --- `test.pgm' near (10,-19).");
    /* Compute the moment matrix, eigenvalues, and edginess: */
    smooth ( subg, g, 10,-19, bump );
    momenta ( mat, subg );
    print2x2( mat, "Left side of edge, x0 = (10,-19):");
    eigenvalues(y,mat);
    printf( "   Eigenvalues: %f > %f > 0;      Edginess %f\n",
	    y[0],y[1], edginess(y) );

    /* MIDDLE, small twisted horizontal edge */
    /* Plot the locale: */
    sharp ( subg, g,  10,3 );    normalize ( subg );
    outfile = fopen ( "testmid.pgm", "wb");  assert(outfile);
    writePGM ( outfile, subg );   fclose ( outfile );
    puts("Wrote testmid.pgm --- `test.pgm' near (10,3).");
    /* Compute the moment matrix, eigenvalues, and edginess: */
    smooth ( subg, g, 10,3, bump );
    momenta ( mat, subg );
    print2x2( mat, "Middle of edge, x0 = (10,3):");
    eigenvalues(y,mat);
    printf( "   Eigenvalues: %f > %f > 0;      Edginess %f\n",
	    y[0],y[1], edginess(y) );

    /* RIGHT, big negative horizontal edge */
    /* Plot the locale: */
    sharp ( subg, g,  10,19 );    normalize ( subg );
    outfile = fopen ( "testrt.pgm", "wb");  assert(outfile);
    writePGM ( outfile, subg );   fclose ( outfile );
    puts("Wrote testrt.pgm --- `test.pgm' near (10,19).");
    /* Compute the moment matrix, eigenvalues, and edginess: */
    smooth ( subg, g, 10,19, bump );
    momenta ( mat, subg );
    print2x2( mat, "Right side of edge, x0 = (10,19):");
    eigenvalues(y,mat);
    printf( "   Eigenvalues: %f > %f > 0;      Edginess %f\n",
	    y[0],y[1], edginess(y) );

   
    /* FLAT, gentle slope, no edge */
    /* Plot the locale: */
    sharp ( subg, g,  -3,-12 );    normalize ( subg );
    outfile = fopen ( "testflat.pgm", "wb");  assert(outfile);
    writePGM ( outfile, subg );   fclose ( outfile );
    puts("Wrote testflat.pgm --- `test.pgm' near (-3,-12).");
    /* Compute the moment matrix, eigenvalues, and edginess: */
    smooth ( subg, g, -3,-12, bump );
    momenta ( mat, subg );
    print2x2( mat, "Gentle slope region, x0 = (-3,-12):");
    eigenvalues(y,mat);
    printf( "   Eigenvalues: %f > %f > 0;      Edginess %f\n",
	    y[0],y[1], edginess(y) );


    freeGrid(bump);
    freeGrid(subg);
  }

  /* Grand Finale: edginess of `test.pgm' */
  {
    Grid edges, subg, bump;
    real emat[2][2], evs[2];
    FILE *outfile;
    int m,n;

    outfile = fopen("testedgi.pgm","wb");
    printf("Writing testedgi.pgm --- edginess plot of `test.pgm'....");
    fflush(stdout);
    edges = likeGrid ( g );
    bump = gaussian(3);
    subg = likeGrid (bump);
    for ( m = g->r0  ;  m <= g->r1  ;  m++ )
      for ( n = g->c0  ;  n <= g->c1  ;  n++ ) {
	smooth ( subg, g, m,n, bump );
	momenta ( emat, subg );
	eigenvalues ( evs, emat );
	insert ( edges, m,n, edginess(evs) );
      }
    normalize ( edges );  writePGM ( outfile, edges);
    fclose(outfile);
    puts("done.");

    freeGrid(edges); freeGrid(subg); freeGrid (bump);
  }
  freeGrid (g);
  fclose(testin);

  return 0;
}
